import numpy as np
import matplotlib.pyplot as plt
import pickle

from systemDynamics import ForwardEuler, Offrice_Area1, Offrice_Area2, Offrice_Area3, Offrice_Area4
from tankControllers import Controllers
from RL_sys import RL_agent

# Parameters from the system dynamics
dt             = 0.01
A1             = Offrice_Area1
A2             = Offrice_Area2
A3             = Offrice_Area3
A4             = Offrice_Area4
H_emergency    = 0.375

#Traditional system
A1_simple      = (0.0075**2) * np.pi
A2_simple      = (0.008**2) * np.pi

Base_system       = False
CONTROL_EXTENSION = False
Advanced_system   = True
Random_switch     = True
emerge1           = False
emerge2           = False

tank2_setpoint = 0.25
ctr = Controllers(dt)

if Base_system or CONTROL_EXTENSION:
    solver = ForwardEuler(dt, A1_simple, A2_simple, A3, A4, emerge1, emerge2)
    setpoint = solver.calc_setpoint(tank2_setpoint)

    A, B, Q, R = ctr.lqr_matrices(setpoint)
    Kp_simple_lqr = ctr.lqr_K_matrix(A, B, Q, R)

if Advanced_system:
    solver = ForwardEuler(dt, A1, A2, A3, A4, emerge1, emerge2)
    setpoint_lqr = solver.calc_setpoint_advanced(tank2_setpoint)
    A_lqr1, B_lqr1, Q_lqr1, R_lqr1 = ctr.lqr_matrices1(setpoint_lqr[0])
    A_lqr2, B_lqr2, Q_lqr2, R_lqr2 = ctr.lqr_matrices1(setpoint_lqr[1])
    A_lqr3, B_lqr3, Q_lqr3, R_lqr3 = ctr.lqr_matrices1(setpoint_lqr[2])
    A_lqr4, B_lqr4, Q_lqr4, R_lqr4 = ctr.lqr_matrices1(setpoint_lqr[3])

    Kp_lqr1 = ctr.lqr_K_matrix(A_lqr1, B_lqr1, Q_lqr1, R_lqr1)
    Kp_lqr2 = ctr.lqr_K_matrix(A_lqr2, B_lqr2, Q_lqr2, R_lqr2)
    Kp_lqr3 = ctr.lqr_K_matrix(A_lqr3, B_lqr3, Q_lqr3, R_lqr3)
    Kp_lqr4 = ctr.lqr_K_matrix(A_lqr4, B_lqr4, Q_lqr4, R_lqr4)


Ki_lqr = [20, 20]
Ki_simple_lqr = [20, 20]

Kp_pi = [500, 500]
Ki_pi = [20, 20]

Kp_pd = [500, 500]
Kd_pd = [0.05, 0.05]

if Base_system:
    u = [ctr.min_controller(solver.u_min), ctr.max_controller(solver.u_max), 0]
    ctr_size = len(u)
if CONTROL_EXTENSION:
    #u = [ctr.min_controller(solver.u_min), ctr.max_controller(solver.u_max), 0, 0, 0]
    u = [0, 0, 0]
    ctr_size = len(u)
if Advanced_system:
    u = [ctr.min_controller(solver.u_min), ctr.max_controller(solver.u_max), 0, 0, 0, 0]
    ctr_size = len(u)

# Discrete states for Q table
h_max        = 0.5
h_min        = 0
step1        = 0.1  #0.0125 = 40 #0.00625 = 80 0.003125 = 160 #0.0015625 = 320
step2        = 0.0125

STATE_SIZE_1 = int((h_max - h_min) / step1)
STATE_SIZE_2 = int((h_max - h_min) / step2)
Easy_state_size = 3

step = [step1, step2]
n            = 20000


# Parameters for the agent
TRAINING = False
SARSA = True
HM_EPISODES = 1
epsilon = 0
gamma = 0.9

start_q_table = ''  #Insert name of Q table file name to get plots
save = False
if HM_EPISODES > 1:
    save = True


#Initialize q_table
if start_q_table is None:
    if Base_system:
        q_table = np.zeros((STATE_SIZE_1 + 1, Easy_state_size, ctr_size))
    if CONTROL_EXTENSION:
        q_table = np.zeros((STATE_SIZE_1 + 1, Easy_state_size, ctr_size))
    if Advanced_system:
        q_table = np.zeros((STATE_SIZE_1 + 1, Easy_state_size, 2, 2, ctr_size))
else:
    with open(start_q_table, 'rb') as f:
        q_table = pickle.load(f)

print(q_table.shape)

#Making agent object
rl = RL_agent(q_table, epsilon, step, gamma, ctr_size)


t = np.linspace(0, dt * n, n)

episode_rewards = []
measure_vec = []


for episode in range(HM_EPISODES):
    episode_reward = 0

    if Base_system or CONTROL_EXTENSION:
        solver.emerge1 = False
        solver.emerge2 = False

    #The system starts with both valves closed
    if Advanced_system:
        solver.emerge1 = False
        solver.emerge2 = False


    height, h, h0 = solver.initialize_dyn(n, tank2_setpoint)

    err = solver.calc_setpoint(tank2_setpoint) - h0

    #Used to start the system at a wanted height
    '''
    h0 = np.array([0.48, 0.25])
    height[:, 0] = h0
    err = solver.calc_setpoint(tank2_setpoint) - height[:, 0]
    #reward = rl.rew_func(h0, h0)
    '''
    if Advanced_system:
        err_lqr1 = setpoint_lqr[0] - h0
        err_lqr2 = setpoint_lqr[1] - h0
        err_lqr3 = setpoint_lqr[2] - h0
        err_lqr4 = setpoint_lqr[3] - h0
        setpoint = setpoint_lqr[0]

    dyn_time_vec = []
    err_vec = []
    err_vec_measure = []
    control_actions = []
    u_vec = []
    u_lqr_vec = []
    emergency_vector = []

    print('Epsilon before episode', episode + 1, rl.epsilon)

    integral_lqr_simple = 0
    integral_lqr1 = 0
    integral_lqr2 = 0
    integral_lqr3 = 0
    integral_lqr4 = 0
    integral_pi = 0
    derivative = 0
    emergency = rl.get_emerge_state(emerge1, emerge2)
    init_disc = rl.get_disc_state(h0, setpoint)

    if Base_system or CONTROL_EXTENSION:
        action = rl.choose_action1(init_disc)
    else:
        action = rl.choose_action2(init_disc, emergency)

    print('Initial height before episode start:', h0)

    for i in range(n):

        # First set up action space with control options)
        cont_max = ctr.max_controller(solver.u_max)
        cont_min = ctr.min_controller(solver.u_min)


        #Choose which system that will run
        if Base_system:
            #The basic system
            integral_lqr_simple += ctr.one_step_integrate(err, u[2])
            cont_lqr = ctr.lqr_Controller(Kp_simple_lqr, err) + np.dot(Ki_simple_lqr, integral_lqr_simple)
            cont_lqr = ctr.cont_restriction(cont_lqr)
            err = setpoint - height[:, i]
            u = [cont_min, cont_max, cont_lqr]

        if CONTROL_EXTENSION:
            #This is the basic system + 2 additional controllers

            integral_lqr_simple += ctr.one_step_integrate(err, u[0])
            integral_pi += ctr.one_step_integrate(err, u[1])
            derivative += ctr.one_step_derivative(err)

            cont_lqr = ctr.lqr_Controller(Kp_simple_lqr, err) + np.dot(Ki_simple_lqr, integral_lqr_simple)
            cont_lqr = ctr.cont_restriction(cont_lqr)
            cont_pi = ctr.P_controller(Kp_pi, err) + np.dot(Ki_pi, integral_pi)
            cont_pi = ctr.cont_restriction(cont_pi)
            cont_pd = ctr.P_controller(Kp_pd, err) + np.dot(Kd_pd, derivative)
            cont_pd = ctr.cont_restriction(cont_pd)

            err = setpoint - height[:, i]

            u = [cont_min, cont_max, cont_lqr, cont_pi, cont_pd]

        if Advanced_system:
            #This is the emergency valve system
            integral_lqr1 += ctr.one_step_integrate(err_lqr1, u[2])
            integral_lqr2 += ctr.one_step_integrate(err_lqr2, u[3])
            integral_lqr3 += ctr.one_step_integrate(err_lqr3, u[4])
            integral_lqr4 += ctr.one_step_integrate(err_lqr4, u[5])

            cont_lqr1 = ctr.lqr_Controller(Kp_lqr1, err_lqr1) + np.dot(Ki_lqr, integral_lqr1)
            cont_lqr2 = ctr.lqr_Controller(Kp_lqr2, err_lqr2) + np.dot(Ki_lqr, integral_lqr2)
            cont_lqr3 = ctr.lqr_Controller(Kp_lqr3, err_lqr3) + np.dot(Ki_lqr, integral_lqr3)
            cont_lqr4 = ctr.lqr_Controller(Kp_lqr4, err_lqr4) + np.dot(Ki_lqr, integral_lqr4)

            cont_lqr1 = ctr.cont_restriction(cont_lqr1)
            cont_lqr2 = ctr.cont_restriction(cont_lqr2)
            cont_lqr3 = ctr.cont_restriction(cont_lqr3)
            cont_lqr4 = ctr.cont_restriction(cont_lqr4)

            err_lqr1 = setpoint_lqr[0] - height[:, i]
            err_lqr2 = setpoint_lqr[1] - height[:, i]
            err_lqr3 = setpoint_lqr[2] - height[:, i]
            err_lqr4 = setpoint_lqr[3] - height[:, i]

            u = [cont_min, cont_max, cont_lqr1, cont_lqr2, cont_lqr3, cont_lqr4]

           #This segment was used to force the valves open/close.
            '''
            if h0[0] >= H_emergency:
                solver.emerge1 = True

            if h0[1] >= H_emergency:
                solver.emerge2 = True
            
            
            if i >= 5000 and i < 10000:
                solver.emerge2 = True
            
            if i >= 5000 and i < 7500:
                solver.emerge2 = False
            if i > 7500:
                solver.emerge1 = True
            '''
            if Random_switch:
                if i >= 0:
                    number1 = 0.5
                    number2 = 0.5
                    if i % 2500 == 0:
                        if np.random.random() > number1:
                            solver.emerge1 = True
                        else:
                            solver.emerge1 = False
                        if np.random.random() > number2:
                            solver.emerge2 = True
                        else:
                            solver.emerge2 = False

            if solver.emerge1 and solver.emerge2:
                setpoint = setpoint_lqr[3]

            elif solver.emerge1:
                setpoint = setpoint_lqr[1]

            elif solver.emerge2:
                setpoint = setpoint_lqr[2]

            else:
                setpoint = setpoint_lqr[0]



        # Get discrete state

        q_state = rl.get_disc_state(h0, setpoint)

        err_measure = (err[1] ** 2) if Base_system or CONTROL_EXTENSION else (err_lqr1[1] ** 2)

        emergency = rl.get_emerge_state(solver.emerge1, solver.emerge2)


        #Next state calculated with the differential equation from solver
        h[:, i] = solver.update(h0, u[action])


        height[:, i + 1] = h[:, i]
        h0 = h[:, i]

        new_obs = height[:, i + 1]


        next_q_state = rl.get_disc_state(new_obs, setpoint)


        if Base_system or CONTROL_EXTENSION:
            next_action = rl.choose_action1(next_q_state)
        else:
            next_action = rl.choose_action2(next_q_state, emergency)

        h_predict = solver.update(h0, u[next_action])



        alpha = rl.alpha_func(i)




        reward, helper = rl.rew_func(height[:, i], h0, h_predict, setpoint)

        #This is used to see the performance of the agent
        if not TRAINING:
            #print('q-table max:', action, q_table[q_state[0], q_state[1], emergency[0], emergency[1], :]) \
                #if Advanced_system \
                #else print('q-table max:', action, q_table[q_state[0], q_state[1], :])
            #print('q-state:', q_state)
            #print('Reward:', reward)
           #print('Integral value', integral)
            print('The value of u:', u)
            print('Last height:', h0)
        #Choose algorithm to run
        if TRAINING:
            if Base_system or CONTROL_EXTENSION:
                if SARSA:
                    new_q = rl.SARSA1(q_state, next_q_state, action, next_action, alpha, reward)
                else:
                    new_q = rl.update_Q1(q_state, next_q_state, action, alpha, reward)
                q_table[q_state[0], q_state[1], action] = new_q
                rl.q_table[q_state[0], q_state[1], action] = new_q
            else:
                if SARSA:
                    new_q = rl.SARSA2(q_state, next_q_state, emergency, action, next_action, alpha, reward)
                else:
                    new_q = rl.update_Q2(q_state, next_q_state, emergency, action, alpha, reward)
                q_table[q_state[0], q_state[1], emergency[0], emergency[1], action] = new_q
                rl.q_table[q_state[0], q_state[1], emergency[0], emergency[1], action] = new_q



        action = next_action



        u_vec.append(u[action])
        u_lqr_vec.append(u[2])
        err_vec.append(err[1]) if Base_system or CONTROL_EXTENSION else err_vec.append(err_lqr1[1])
        err_vec_measure.append(err_measure)
        control_actions.append(action)
        dyn_time_vec.append(i)


        if solver.emerge1 and solver.emerge2:
            emergency_vector.append(4)
        elif solver.emerge1:
            emergency_vector.append(2)
        elif solver.emerge2:
            emergency_vector.append(3)
        else:
            emergency_vector.append(1)

    episode_rewards.append(episode_reward)
    measure_metric = sum(err_vec_measure)
    measure_vec.append(measure_metric)


    print('Measure metric:', measure_metric)


    if rl.epsilon >= 0.1:
        rl.epsilon *= rl.eps_decay

    if episode == HM_EPISODES - 2 and HM_EPISODES > 100:
        rl.epsilon = 0




avg_ep_rew = np.sum(episode_rewards) / len(episode_rewards)



episodes = np.linspace(0, HM_EPISODES, HM_EPISODES)
t_sim = np.linspace(0, n, n + 1)

plt.figure(1)
plt.rcParams.update({'font.size': 22})
plt.title('Control actions')
plt.plot(t_sim[:-1], control_actions)
plt.xlabel('Timestep [n]')
plt.ylabel('Actions')
plt.grid(True)

'''
plt.figure(5)
plt.title('Lqr values')
plt.plot(t_sim[:-1], u_lqr_vec)
plt.xlabel('Timestep [n]')
plt.ylabel('Value')
plt.grid(True)

plt.figure(2)
plt.title('Error from setpoint in tank 2')
plt.plot(t_sim[:-1], err_vec, label='error tank 2 in episode:' + str(episode + 1))
plt.xlabel('Timestep [n]')
plt.ylabel('Height [m]')
plt.legend(loc='best')
plt.grid(True)
'''
plt.figure(3)
plt.title('Height in both tanks')
plt.plot(t_sim, height[0, :], 'r', label='height tank 1 in episode:' + str(episode + 1))
plt.plot(t_sim, height[1, :], 'g', label='height tank 2 in episode: ' + str(episode + 1))
#plt.hlines(setpoint[0], 0, n, colors='y', linestyles='--', label='Equilibrium for tank 1')
#plt.hlines(setpoint[1], 0, n, colors='y', linestyles='--', label='Equilibrium for tank 2')
plt.hlines(setpoint_lqr[0], 0, n, colors='y', linestyles='--', label='Equilibrium for subsystem 1')
plt.hlines(setpoint_lqr[1], 0, n, colors='m', linestyles='--', label='Equilibrium for subsystem 2')
plt.hlines(setpoint_lqr[2], 0, n, colors='c', linestyles='--', label='Equilibrium for subsystem 3')
plt.hlines(setpoint_lqr[3], 0, n, colors='k', linestyles='--', label='Equilibrium for subsystem 4')
plt.hlines(setpoint[0], 0, n, colors='m', linestyles='--')
plt.hlines(setpoint[1], 0, n, colors='b', linestyles='--')
plt.xlabel('Timestep [n]')
plt.ylabel('Height [m]')
plt.grid(True)
plt.legend(loc='best')

plt.figure(6)
plt.title('Emergency valves')
plt.plot(t_sim[:-1], emergency_vector)
plt.xlabel('Timesteps [n]')
plt.ylabel('Emergency subsystem')
plt.grid(True)

if HM_EPISODES != 1:
    plt.figure(4)
    plt.title('Meassure metric')
    plt.plot(episodes, measure_vec)
    plt.xlabel('Episodes')
    plt.ylabel('err^2')


plt.show()

if save:
    with open(f"{HM_EPISODES}_ep + {helper}.pickle", "wb") as f:
        pickle.dump(q_table, f)
