import numpy as np






class RL_agent(object):
    def __init__(self, q_table, epsilon, step, gamma, ctr_size):
        self.q_table, self.epsilon, self.step, self.gamma, self.ctr_size = q_table, epsilon, step, gamma, ctr_size
        self.eps_decay = 0.995




    #Action selection for the basesystem
    def choose_action1(self, state):
        if np.random.random() > self.epsilon:
            action = np.argmax(self.q_table[state[0], state[1], :])
        else:
            action = np.random.randint(0, self.ctr_size)
        return action

    #Action selection for the emergency valve system
    def choose_action2(self, state, emergency):
        if np.random.random() > self.epsilon:
            action = np.argmax(self.q_table[state[0], state[1], emergency[0], emergency[1], :])
        else:
            action = np.random.randint(0, self.ctr_size)
        return action



    def get_disc_state(self, h, setpoint):

        h1_disc = int(round(h[0] / self.step[0]))
        h2_disc = int(round(h[1] / self.step[1]))

        setpoint_disc = int(round(setpoint[1] / self.step[1]))


        if h2_disc > setpoint_disc:
            state2_disc = 2
        elif h2_disc == setpoint_disc:
            state2_disc = 1
        else:
            state2_disc = 0


        state1_disc = h1_disc
        #state2_disc = h2_disc
        disc_state_vector = [state1_disc, state2_disc]

        return disc_state_vector


    def get_emerge_state(self, emerge1, emerge2):
        emergency = [0, 0]
        if emerge1:
            emergency[0] = 1
        if emerge2:
            emergency[1] = 1

        return emergency


    def alpha_func(self, n):

        alpha = 1 / (1 + (n*0.0125))
        return alpha



    def rew_func(self, h, h_next, h_pred, setpoint):
        err = setpoint - h
        err_next = setpoint - h_next
        err_pred = setpoint - h_pred

        #Helper to name q-tables
        helper = ''                     #Used to seperate the Q table names saved for testing

        reward = - ((err_pred[1] ** 2 + err_next[1] ** 2))

        return reward, helper

    #Q learning update rule
    def update_Q1(self, state, n_state, action, alpha, reward):

        max_future_q = np.max(self.q_table[n_state[0], n_state[1], :])

        current_q = self.q_table[state[0], state[1], action]

        new_q = ((1 - alpha) * current_q) + (alpha * (reward + (self.gamma * max_future_q)))

        return new_q


    #Q learning update rule with emergency states
    def update_Q2(self, state, n_state, emergency, action, alpha, reward):

        max_future_q = np.max(self.q_table[n_state[0], n_state[1], emergency[0], emergency[1], :])

        current_q = self.q_table[state[0], state[1], emergency[0], emergency[1], action]

        new_q = ((1 - alpha) * current_q) + (alpha * (reward + (self.gamma * max_future_q)))

        return new_q


    #SARSA update rule
    def SARSA1(self, state, n_state, action, action2, alpha, reward):

        next_q = self.q_table[n_state[0], n_state[1], action2]
        current_q = self.q_table[state[0], state[1], action]

        new_q = ((1 - alpha) * current_q) + (alpha * (reward + (self.gamma * next_q)))

        return new_q


    #SARSA update rule with emergency states
    def SARSA2(self, state, n_state, emergency, action, action2, alpha, reward):

        next_q = self.q_table[n_state[0], n_state[1], emergency[0], emergency[1], action2]
        current_q = self.q_table[state[0], state[1], emergency[0], emergency[1], action]

        new_q = ((1 - alpha) * current_q) + (alpha * (reward + (self.gamma * next_q)))

        return new_q
