import numpy as np
import random
import pickle



#Constants
'Base system had r1 = 0.0075 and r2 = 0.008'

G = 9.81
A_tank1       = (0.05**2) * np.pi
A_tank2       = (0.05**2) * np.pi                       #Area of both tanks
Offrice_Area1 = (0.006**2) * np.pi                     #Area of outlet from tank 1
Offrice_Area2 = (0.005**2) * np.pi                      #Area of outlet from tank 2
Offrice_Area3 = (0.004**2) * np.pi
Offrice_Area4 = (0.003**2) * np.pi
C             = (2*G)**0.5                              #Constant from outlet 1
K             = 0.0002                                  #Pump constant





class ForwardEuler(object):
    def __init__(self, dt, A1, A2, A3, A4, emerge1, emerge2):
        self.dt, self.A1, self.A2, self.A3, self.A4 = dt, A1, A2, A3, A4
        self.emerge1 = emerge1
        self.emerge2 = emerge2
        self.u_max = 16
        self.u_min = -10


    def outlet(self, h):
        leakage = 0.000001

        c1 = self.A1 * C + np.abs((leakage * np.sin(2 * np.pi * h[0])) ** 2) * np.pi
        c2 = self.A2 * C + np.abs((leakage * np.sin(2 * np.pi * h[1])) ** 2) * np.pi

        emergency_c1 = self.A3 * C + np.abs((leakage * np.sin(2 * np.pi * h[0])) ** 2) * np.pi
        emergency_c2 = self.A4 * C + np.abs((leakage * np.sin(2 * np.pi * h[0])) ** 2) * np.pi

        return c1, c2, emergency_c1, emergency_c2

    def calc_setpoint_advanced(self, tank2_setpoint):

        out1 = Offrice_Area1 * C
        out2 = Offrice_Area2 * C
        out3 = Offrice_Area3 * C
        out4 = Offrice_Area4 * C

        setpoint1 = [((out2 / out1) * np.sqrt(tank2_setpoint)) ** 2, tank2_setpoint]

        setpoint2 = [((out2 / (out1 + out3)) * np.sqrt(tank2_setpoint)) ** 2, tank2_setpoint]

        setpoint3 = [(((out2 + out4) / out1) * np.sqrt(tank2_setpoint)) ** 2, tank2_setpoint]

        setpoint4 = [(((out2 + out4) / (out1 + out3)) * np.sqrt(tank2_setpoint)) ** 2, tank2_setpoint]

        setpoint = [setpoint1, setpoint2, setpoint3, setpoint4]

        return setpoint


    def calc_setpoint(self, tank2_setpoint):
        A1_simple = (0.0075 ** 2) * np.pi
        A2_simple = (0.008 ** 2) * np.pi

        out1 = A1_simple * C
        out2 = A2_simple * C

        setpoint = [((out2 / out1) * np.sqrt(tank2_setpoint)) ** 2, tank2_setpoint]

        return setpoint


    def initialize_dyn(self, n, tank2_setpoint):
        height = np.zeros((2, n + 1))
        h0 = [random.random(), random.random()]

        if h0[0] > 0.5:
            h0[0] = h0[0] - 0.5
        if h0[1] > 0.5:
            h0[1] = h0[1] - 0.5

        height[:, 0] = h0
        h = np.zeros((2, n + 1))

        return height, h, np.array(h0)



    def f(self, x, u):
        #2 tank dynamics
        x1 = x[0]
        x2 = x[1]



        oc1, oc2, emergency_c1, emergency_c2 = self.outlet(x)

        #Inflow to tank1:
        qin = K * u

        #Outflow of tank 1
        qout1_tank1 = oc1 * np.sqrt(x1)
        qout2_tank1 = 0

        #Outflow of tank 2
        qout1_tank2 = oc2 * np.sqrt(x2)
        qout2_tank2 = 0



        if self.emerge1:
            qout2_tank1 = emergency_c1 * np.sqrt(x1)

        if self.emerge2:
            qout2_tank2 = emergency_c2 * np.sqrt(x2)

        dx1dt = (qin - qout1_tank1 - qout2_tank1) / A_tank1
        dx2dt = (qout1_tank1 + qout2_tank1 - qout1_tank2 - qout2_tank2) / A_tank2

        dxdt = [dx1dt, dx2dt]

        return np.array(dxdt)




    def eulerUpdate(self, h, u):
        dt = self.dt,
        f = self.f(h, u)


        deltaH = dt * f

        return deltaH


    def update(self, h, u):


        deltaH = self.eulerUpdate(h, u)

        newState = h + deltaH

        #Overflow conditions

        if newState[0] >= 0.5 and deltaH[0] > 0:
            newState[0] = h[0]
        if newState[1] >= 0.5 and deltaH[1] > 0:
            newState[1] = h[1]


        #Underflow conditions

        if newState[0] < 0.0 and deltaH[0] < 0:

            newState[0] = 0
        if newState[1] < 0.0 and deltaH[1] < 0:

            newState[1] = 0

        return newState









