import numpy as np
import scipy.linalg
from systemDynamics import C, Offrice_Area1, Offrice_Area2, Offrice_Area3, Offrice_Area4, A_tank1, A_tank2, K






class Controllers(object):
    def __init__(self, dt):
        self.dt = dt
        self.lqr_max = 8
        self.lqr_min = -5





    def lqr_matrices(self, setpoint):
        # Parameters for original system

        c1 = ((0.0075**2) *np.pi) * C
        c2 = ((0.008**2) *np.pi) * C
        a11 = -(c1 / (A_tank1 * (2 * np.sqrt(setpoint[0]))))
        a12 = 0
        a21 = -a11
        a22 = -(c2 / (A_tank2 * (2 * np.sqrt(setpoint[1]))))

        A = np.array([[a11, a12], [a21, a22]])
        B = np.array([[K / A_tank1], [0]])
        Q = np.array([[6000000, 0], [0, 24000000]])
        R = 1

        return A, B, Q, R


    def lqr_matrices1(self, setpoint):
        #setpoint = [0.12056327, 0.25]

        c1 = Offrice_Area1 * C
        c2 = Offrice_Area2 * C
        a11 = -(c1 / (A_tank1 * (2 * np.sqrt(setpoint[0]))))
        a12 = 0
        a21 = (c1) / (2 * A_tank2 * np.sqrt(setpoint[0]))
        a22 = -(c2 / (A_tank2 * (2 * np.sqrt(setpoint[1]))))

        A = np.array([[a11, a12], [a21, a22]])
        B = np.array([[K / A_tank1], [0]])
        Q = np.array([[6000000, 0], [0, 24000000]])
        R = 1


        return A, B, Q, R



    def lqr_matrices2(self, setpoint):
        #setpoint = [0.057785, 0.25]

        c1 = Offrice_Area1 * C
        c2 = Offrice_Area2 * C
        c3 = Offrice_Area3 * C

        a11 = -((c1 + c3) / (A_tank1 * (2 * np.sqrt(setpoint[0]))))
        a12 = 0
        a21 = (c1 + c3) / (2 * A_tank2 * np.sqrt(setpoint[0]))
        a22 = -(c2 / (A_tank2 * (2 * np.sqrt(setpoint[1]))))

        A = np.array([[a11, a12], [a21, a22]])
        B = np.array([[K / A_tank1], [0]])
        Q = np.array([[6000000, 0], [0, 24000000]])
        R = 1

        return A, B, Q, R


    def lqr_matrices3(self, setpoint):
        #setpoint = [0.222994, 0.25]

        c1 = Offrice_Area1 * C
        c2 = Offrice_Area2 * C
        c4 = Offrice_Area4 * C
        a11 = -(c1 / (A_tank1 * (2 * np.sqrt(setpoint[0]))))
        a12 = 0
        a21 = (c1) / (2 * A_tank2 * np.sqrt(setpoint[0]))
        a22 = -((c2 + c4) / (A_tank2 * (2 * np.sqrt(setpoint[1]))))

        A = np.array([[a11, a12], [a21, a22]])
        B = np.array([[K / A_tank1], [0]])
        Q = np.array([[6000000, 0], [0, 24000000]])
        R = 1

        return A, B, Q, R


    def lqr_matrices4(self, setpoint):
        #setpoint = [0.10688, 0.25]

        c1 = Offrice_Area1 * C
        c2 = Offrice_Area2 * C
        c3 = Offrice_Area3 * C
        c4 = Offrice_Area4 * C
        a11 = -((c1 + c3) / (A_tank1 * (2 * np.sqrt(setpoint[0]))))
        a12 = 0
        a21 = (c1 + c3) / (2 * A_tank2 * np.sqrt(setpoint[0]))
        a22 = -((c2 + c4) / (A_tank2 * (2 * np.sqrt(setpoint[1]))))

        A = np.array([[a11, a12], [a21, a22]])
        B = np.array([[K / A_tank1], [0]])
        Q = np.array([[6000000, 0], [0, 24000000]])
        R = 1

        return A, B, Q, R



    def lqr_K_matrix(self, A, B, Q, R):
        X = scipy.linalg.solve_continuous_are(A, B, Q, R)
        K = np.dot(1/R, (np.dot(B.T, X)))
        K1 = K[0, 0]
        K2 = K[0, 1]
        K_lqr  = [K1, K2]
        return np.array(K_lqr)


    def max_controller(self, max):
        return max


    def min_controller(self, min):
        return min



    def one_step_integrate(self, e, u):

        if u < self.lqr_max and u > self.lqr_min:
            integral = e * self.dt
        else:
            integral = np.array([float(0), float(0)])

        return integral


    def one_step_derivative(self, e):

        derivative = e / self.dt

        return derivative


    def P_controller(self, K_p, e):

        proportional = np.dot(K_p, e)

        return proportional


    def lqr_Controller(self, K_lqr, e):


        u = np.dot(K_lqr, e)

        return u


    def cont_restriction(self, u):

        if u > self.lqr_max:
            return self.lqr_max
        elif u < self.lqr_min:
            return self.lqr_min
        else:
            return u
